/*
 * A pipe paging coordinator, permitting memory sharing pipe endpoints.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <algorithm>

#include <mem/memory_preallocated.h>

#include "page_queue_partitioned.h"
#include "pipe_paging.h"



PipePaging::PipePaging(Memory *memory, offset_t size)
: NotificationSupport(2), _memory(NULL)
{
    /* Employ a region size compatible with the configured page size. */

    _size = std::max<offset_t>(size, PAGE_SIZE);

    /* Reserve space for two pipe regions. */

    _memory = new MemoryPreallocated(memory, round(size, memory->region_size()) * 2);
    _queue = new PageQueuePartitioned();
    _pages = new Pages(_memory, _queue);

    /* Reset the mappers associated with the regions. */

    for (unsigned int i = 0; i < 2; i++)
        _regions[i] = NULL;

    /* Initialise a region accessible by reader and writer. */

    _add_region();
    _next_region();
}

PipePaging::~PipePaging()
{
    /* Discard all regions from the pipe. */

    for (unsigned int i = 0; i < 2; i++)
        discard_region(i);

    /* Release notifiers. */

    release_notifiers();

    /* Delete the page collection and related objects. */

    delete _pages;
    delete _queue;
    delete _memory;
}

/* Return whether one or more endpoints have detached. */

int PipePaging::closed()
{
    return _active_endpoints < 2;
}

/* Detach one endpoint, returning the number still active. */

unsigned int PipePaging::detach()
{
    std::lock_guard<std::mutex> guard(_lock);

    if (_active_endpoints)
        _active_endpoints--;

    return _active_endpoints;
}

/* Discard a region. */

void PipePaging::discard_region(unsigned int i)
{
    PageMapper *mapper = _regions[i];

    if (mapper != NULL)
    {
        _regions[i] = NULL;
        delete mapper;
    }
}

/* Add a region to the sequence. */

PageMapper *PipePaging::_add_region()
{
    /* If the writer already accesses a different region to the reader, no new
       region is added. */

    if (_writing != _reading)
        return NULL;

    /* Select the other region of the pair being maintained. */

    _writing = 1 - _writing;

    /* Make a new mapper for the region. */

    PageMapper *mapper = new PageMapper(&_accessors[_writing], _pages);

    /* Initialise and record the mapper. */

    mapper->set_data_size(0);

    _regions[_writing] = mapper;

    /* Return the next region's mapper. */

    return mapper;
}

PageMapper *PipePaging::add_region()
{
    PageMapper *mapper = _add_region();

    /* Let the writer notify the reader. */

    notify_others(PipePaging::WRITER, NOTIFY_CONTENT_AVAILABLE, NOTIFY_VALUES_NULL);

    /* Return the next region's mapper. */

    return mapper;
}

/* Return the current region for reading. */

PageMapper *PipePaging::current_region()
{
    return _regions[_reading];
}

/* Return the current region for reading or writing. */

PageMapper *PipePaging::current_region(bool writing)
{
    return _regions[writing ? _writing : _reading];
}

/* Return the next region for the reader if the writer is using a different one.
   Otherwise, return NULL. */

PageMapper *PipePaging::_next_region()
{
    /* If the reader already accesses the same region as the writer, no next
       region can be obtained. */

    if (_reading == _writing)
        return NULL;

    /* Detach and discard the current page mapper. */

    discard_region(_reading);

    /* Select the next region. */

    _reading = 1 - _reading;

    /* Return the next region's mapper. */

    return _regions[_reading];
}

PageMapper *PipePaging::next_region()
{
    PageMapper *mapper = _next_region();

    /* Let the reader notify the writer. */

    notify_others(PipePaging::READER, NOTIFY_SPACE_AVAILABLE, NOTIFY_VALUES_NULL);

    return mapper;
}

// vim: tabstop=4 expandtab shiftwidth=4
